@@ -23,10 +23,14 @@ from scipy.ndimage.filters import gaussian_filter
 from scipy.interpolate import UnivariateSpline as uspline
 from scipy.integrate import simpson
 
-from dask.distributed import Client,LocalCluster
-import dask.array as da
-
+from functools import wraps
 
+USE_DASK = True
+try:
+    from dask.distributed import Client,LocalCluster
+    import dask.array as da
+except:
+    USE_DASK = False
 
 def lsh5(hd, prefix='', top_only=False, silent=False, print_attrs=True):
     """ list the content of a HDF5 file
@@ -64,6 +68,54 @@ def create_linked_files(fn, fnlist):
         fs.close()
     ff.close()
 
+def get_det_timestamps(grp, trigger, exp, err_band=5., debug=False):
+    """ in fly scans, the timestamps comes from the XPS controller. sometimes the timestamps appear to be off
+        the timestamp on the detector image is used as a fallback
+        
+        also check the length of the timestamp data, the total number of data points should agree of 
+        number of frames in the scattering data
+    """
+    ts0 = grp[f'timestamps/{trigger}'][...].flatten()
+    dns = [dn for dn in grp['timestamps'].keys() if 'image' in dn]
+    if len(dns)<1:
+        raise Exception("cannot find detector data in {grp}")
+    tsd = grp[f'timestamps/{dns[0]}'][...]
+
+    dshape = grp[f'data/{dns[0]}'].shape[:-2]
+    if len(dshape)>1:
+        if len(dshape)>2:
+            raise Exception(f"Don't know how to handle data shape {dshape}")
+        nimg = dshape[0]*dshape[1]
+        ts0 = ts0.flatten()
+    else:
+        nimg = dshape[0]
+    if len(ts0) != nimg:
+        print(f"Warning: mistached timestamp length: {len(ts0)} vs {nimg}")
+        #raise Exception(f"mistached timestamp length: {len(ts0)} vs {dshape}")
+    
+    if debug:
+        print(f"time stamp from {trigger} ends at {ts0[-1]} .")
+        print(f"time stamp from {dns[0]} ends at {tsd[-1]} .")
+        print(f"off by {ts0[-1]-tsd[-1]:.3f}")
+        print(f"size of the timestamp data: {len(ts0)}")
+        print(f"size of the scattering data : {nimg}")
+    
+    if len(ts0)<nimg:
+        print("Warning: more scattering patterns than recorded triggers, repeating the last triggers ...")
+        ts01 = ts0[:nimg-len(ts0)]
+        ts0 = np.append(ts0, ts01-ts01[0]+ts0[-1]+exp)
+        #raise Exception("more scattering patterns than recorded triggers ...")
+    elif len(ts0)>nimg:
+        print("Warning: more triggers than existing scattering patterns, ignore the extra triggers ... ")
+        ts0 = ts0[:nimg]
+    
+    dt = ts0[-1]-tsd[-1]
+    if np.fabs(dt)>err_band:
+        print(f"Warning: strange trigger time stamps (different by {dt:.2f} s), falling back to {dns[0]} data")
+        ts0 += tsd[-1]-exp-ts0[-1]
+
+    return ts0    
+    
 def synch_mon(ts0, em0, ts, em, dt="auto", Ns=8, plot=False):
     """ revise the timestamp ts (np array) to match the monitor em to em0
         look for the most significant change, i.e. when the shutter opens 
@@ -146,7 +198,8 @@ def get_monitor_counts(grp, monitorName):
             if monitorName in fieldName:
                 strn = stream
                 data = grp[strn]["data"][fieldName][...]
-                ts = grp[strn]["timestamps"][fieldName][...]
+                #ts = grp[strn]["timestamps"][fieldName][...]   # this filed seems to have occasional issues 
+                ts = grp[strn]["time"][...]   # this is consistent with data from databroker
                 break
     if strn is None:
         raise Exception(f"could not find the stream that contains {monitorName}.")
@@ -321,7 +374,7 @@ class h5exp():
         self.fh5.close()
         return np.asarray(qgrid)
     
-    def recalibrate(self, fn_std, energy=-1, e_range=[5, 20], use_recalib=False,
+    def recalibrate(self, fn_std, energy, e_range=[5, 20], use_recalib=False,
                     det_type={"_SAXS": "Pilatus1M", "_WAXS2": "Pilatus1M"},
                     bkg={}, temp_file_location="/tmp"):
         """ fn_std should be a h5 file that contains AgBH pattern
@@ -336,30 +389,32 @@ class h5exp():
             wl = 2.*np.pi*1.973/energy
             for det in self.detectors:
                 det.exp_para.wavelength = wl
-        elif energy>0:
+        else:
             raise Exception(f"energy should be between {e_range[0]} and {e_range[1]} (keV): {energy}")
         
         for det in self.detectors:
             print(f"processing detector {det.extension} ...")    
             ep = det.exp_para
             data_file = f"{temp_file_location}/{uname}{det.extension}.cbf"
-            img = dstd.fh5["%s/primary/data/%s" % (sn, dstd.det_name[det.extension])][0]
 
-            # this would work better if the detector geometry specification 
-            # can be more flexible for pyFAI-recalib 
-            if ep.flip: ## can only handle flip=1 right now
-                if ep.flip!=1: 
-                    raise Exception(f"don't know how to handle flip={ep.flip}.")
-                poni1 = pxsize*ep.bm_ctr_x
-                poni2 = pxsize*(ep.ImageHeight-ep.bm_ctr_y)
-                dmask = np.fliplr(det.exp_para.mask.map.T)
-            else: 
-                poni1 = pxsize*ep.bm_ctr_y
-                poni2 = pxsize*ep.bm_ctr_x
-                dmask = det.exp_para.mask.map
-            if det.extension in bkg.keys():
-                img -= bkg[det.extension]
-            fabio.cbfimage.CbfImage(data=img*(~dmask)).write(data_file)
+            with h5py.File(dstd.fn, "r") as fh5:
+                img = fh5["%s/primary/data/%s" % (sn, dstd.det_name[det.extension])][0]
+
+                # this would work better if the detector geometry specification 
+                # can be more flexible for pyFAI-recalib 
+                if ep.flip: ## can only handle flip=1 right now
+                    if ep.flip!=1: 
+                        raise Exception(f"don't know how to handle flip={ep.flip}.")
+                    poni1 = pxsize*ep.bm_ctr_x
+                    poni2 = pxsize*(ep.ImageHeight-ep.bm_ctr_y)
+                    dmask = np.fliplr(det.exp_para.mask.map.T)
+                else: 
+                    poni1 = pxsize*ep.bm_ctr_y
+                    poni2 = pxsize*ep.bm_ctr_x
+                    dmask = det.exp_para.mask.map
+                if det.extension in bkg.keys():
+                    img -= bkg[det.extension]
+                fabio.cbfimage.CbfImage(data=img*(~dmask)).write(data_file)
 
             if use_recalib:
                 # WARNING: pyFAI-recalib is obselete
@@ -444,13 +499,35 @@ def find_field(fh5, fieldName, sname=None):
         return tstream
     
     return tstream[sname] 
+
+
+def h5_file_access(method):
+    @wraps(method)
+    def inner(ref, *args, **kwargs):
         
+        if not ref.fh5:
+            file_open = False
+            ref.fh5 = h5py.File(ref.fn, "r", swmr=True)
+        else:
+            file_open = True
+            
+        try:
+            ret = method(ref, *args, **kwargs)
+            if not file_open:  # leave the file open if it was initially open
+                ref.fh5.close()
+            return ret
+        except:
+            ref.fh5.close()
+            raise
+    
+    return inner
+    
 class h5xs():
     """ Scattering data in transmission geometry
         Transmitted beam intensity can be set either from the water peak (sol), or from intensity monitor.
         Data processing can be done either in series, or in parallel. Serial processing can be forced.
         
-        keep a h5 file handle for reading, self.fh5
+        open h5 file only when necessary, using the h5_file_access decorator
         re-open the file for writing only when necessary
             save_attributes??
             save_detectors(), save_d1s()
@@ -472,24 +549,42 @@ class h5xs():
         self.transField = ''  
         self.transStream = {} 
         self.read_only = read_only
-        self.writable = False   # this is the current state of th eh5 file
+        self.writable = False   # this is the current state of the h5 file
 
         self.fn = fn
-        self.fh5 = h5py.File(fn, "r", swmr=True)   # file must exist; reopen for writing only when necessary
+        self.fh5 = None
+        #self.fh5 = h5py.File(fn, "r", swmr=True)   # file must exist; reopen for writing only when necessary
+        self.save_d1 = save_d1
+
+        self.setup(exp_setup, transField, have_raw_data)
+        
+    def explict_open_h5(self):
+        if self.fh5:
+            print(f"{self.hn} file is already open.")
+        else:
+            self.fh5 = h5py.File(self.fn, "r")
+            
+    def explicit_close_h5(self):
+        if not self.fh5:
+            print(f"{self.hn} file is already closed.")
+        else:
+            self.fh5.close()
+        
+    @h5_file_access  
+    def setup(self, exp_setup=None, transField=transmitted_monitor, have_raw_data=True):
         if exp_setup is None:     # assume the h5 file will provide the detector config
             self.qgrid = self.read_detectors()
         else:
             self.detectors, self.qgrid = exp_setup
-            if not read_only:
+            if not self.read_only:
                 self.save_detectors()
         self.list_samples(quiet=True)
 
         if have_raw_data:
             # find out what are the fields corresponding to the 2D detectors
             # at LiX there are two possibilities; assume all samples have have data stored in the same fileds
-            self.save_d1 = save_d1
             sn = self.samples[0]
-            streams = list(self.fh5[f"{sn}"])
+            streams = list(self.fh5[sn].keys())
             data_fields = {}
             for stnm in streams:
                 if 'data' in self.fh5[f"{sn}/{stnm}"].keys(): 
@@ -536,7 +631,7 @@ class h5xs():
                 self.transField = transField
                 self.transMode = trans_mode.external
 
-            if not read_only:
+            if not self.read_only:
                 self.enable_write(True)
                 self.fh5.attrs['trans'] = ','.join([str(self.transMode.value), self.transField])  #elf.transStream])
                 self.enable_write(False)
@@ -562,6 +657,7 @@ class h5xs():
                 return d
         return None
     
+    @h5_file_access  
     def save_detectors(self):
         if self.read_only:
             return
@@ -571,21 +667,45 @@ class h5xs():
         self.fh5.attrs['qgrid'] = list(self.qgrid)
         self.enable_write(False)
     
+    @h5_file_access  
     def read_detectors(self):
         dets_attr = self.fh5.attrs['detectors']
         qgrid = self.fh5.attrs['qgrid']
         self.detectors = [create_det_from_attrs(attrs) for attrs in json.loads(dets_attr)]  
         return np.asarray(qgrid)
 
-    def dset(self, dsname, data_type="data", sn=None, get_path=False):
+    @h5_file_access
+    def get_h5_attr(self, grp, attr_name):
+        return self.fh5[grp].attrs[attr_name]
+
+    @h5_file_access
+    def set_h5_attr(self, grp, attr_name, value):
+        self.enable_write(True)
+        self.fh5[grp].attrs[attr_name] = value
+        self.enable_write(False)
+        
+    @h5_file_access
+    def dshape(self, dsname, data_type="data", sn=None):
+        assert (data_type in ["data", "timestamps"])
+        if sn is None:
+            sn = self.samples[0]
+        strm = find_field(self.fh5, dsname, sn)
+        return self.fh5[f"{sn}/{strm}/{data_type}/{dsname}"].shape
+    
+    @h5_file_access
+    def dset(self, dsname, data_type="data", sn=None, get_path=False, return_reference=True):
         assert (data_type in ["data", "timestamps"])
         if sn is None:
             sn = self.samples[0]
         strm = find_field(self.fh5, dsname, sn)
         if get_path:
             return f"{sn}/{strm}/{data_type}/{dsname}"
-        return self.fh5[f"{sn}/{strm}/{data_type}/{dsname}"]
+        if return_reference:
+            return self.fh5[f"{sn}/{strm}/{data_type}/{dsname}"]   # just return a reference  
+        else:
+            return self.fh5[f"{sn}/{strm}/{data_type}/{dsname}"][...]
     
+    @h5_file_access  
     def md_dict(self, sn, md_keys=[]):
         """ create the meta data to be recorded in ascii data files
             from the detector_config (attribute of the h5xs object) 
@@ -658,18 +778,21 @@ class h5xs():
         
         return md_str
     
+    @h5_file_access  
     def header(self, sn):
         if not sn in self.samples:
-            raise Exception(f"{sn} is not a valie sample.")
+            raise Exception(f"{sn} is not a valid sample.")
         if "start" in self.fh5[sn].attrs:
             return json.loads(self.fh5[sn].attrs['start'])
         return None
     
+    @h5_file_access  
     def list_samples(self, quiet=False):
         self.samples = lsh5(self.fh5, top_only=True, silent=True)
         if not quiet:
             print(self.samples)
     
+    @h5_file_access  
     def verify_frn(self, sn, frn, flatten=False):
         """ simply translate between a scaler index and a multi-dimensional index based on scan shape
             this became more complicated when areadetector saves data as hdf
@@ -720,6 +843,7 @@ class h5xs():
         frn = self.verify_frn(sn, frn, flatten=True)
         return self.d1s[sn][group][frn]    
             
+    @h5_file_access  
     def get_d2(self, sn=None, det_ext=None, frn=None, dtype=None, detectors=None, client=None):
         if sn is None:
             sn = self.samples[0]
@@ -736,6 +860,8 @@ class h5xs():
             except:
                 continue
             if frn=="average":
+                if not USE_DASK:
+                    raise Exception("Dask is not available for averaging frames ...")
                 imgs = da.array(dset)
                 fut = da.average(imgs, axis=tuple(range(len(dset.shape)-2)))
                 davg = fut.compute(client=client)
@@ -881,8 +1007,24 @@ class h5xs():
                               logScale=logScale, useMask=useMask, clim=clim, 
                               aspect=aspect, cmap=cmap, colorbar=colorbar, **kwargs)
 
+    @h5_file_access  
+    def get_ts(self, sn, exp, trigger="sol"):
+        if trigger=="sol" or trigger is None:
+            dn = list(self.det_name.values())[0]
+            # expect a finite but minimal offset in time since all come from the same IOC server
+            ts0 = self.fh5[f'{sn}/primary/timestamps/{dn}'][...].flatten()
+            dshape = self.fh5[f"{sn}/primary/data/{dn}"].shape[0]   # length of the time sequence
+            if len(ts0)==1: # multiple exposures, single trigger, as in HT measurements
+                ts0 = ts0[0]+np.arange(dshape)*exp    
+        elif trigger in self.fh5[f'{sn}/primary/timestamps'].keys():
+            ts0 = get_det_timestamps(self.fh5[f'{sn}/primary/'], trigger, exp)
+        else:
+            raise Exception(f"timestamp data for {trigger} cannot be found.")
+
+        return ts0
     
-    def get_mon(self, sn=None, trigger=None, gf_sigma=2, exp=1, 
+    @h5_file_access  
+    def get_mon(self, sn=None, trigger=None, gf_sigma=2, exp=1, check_size=0,
                 force_synch='auto', force_synch_trig=0, extend_mon_stream=True,
                 plot_trigger=False, **kwargs): 
         """ calculate the monitor counts for each data point
@@ -922,38 +1064,16 @@ class h5xs():
             # expect trans and incid monitor data to be in the same stream
             strn,ts2,trans_data = get_monitor_counts(self.fh5[sn], transmitted_monitor)
             strn,ts1,incid_data = get_monitor_counts(self.fh5[sn], incident_monitor)
-            # synch em1 time stamps to em2, LiX-specific 
-            synch_mon(ts2, trans_data, ts1, incid_data, dt=force_synch)
                 
             if strn=="primary":
                 print("monitors are used as detectors.")
-            elif trigger=="sol" or trigger is None:
-                dn = list(self.det_name.values())[0]
-                # expect a finite but minimal offset in time since all come from the same IOC server
-                ts0 = self.fh5[f'{sn}/primary/timestamps/{dn}'][...].flatten()
-                dshape = self.fh5[f"{sn}/primary/data/{dn}"].shape[0]   # length of the time sequence
-                if len(ts0)==1: # multiple exposures, single trigger, as in HT measurements
-                    ts0 = ts0[0]+np.arange(dshape)*exp    
-            elif trigger in self.fh5[f'{sn}/primary/timestamps'].keys():
-                ts0 = self.fh5[f'{sn}/primary/timestamps/{trigger}'][...].flatten()
-                dshape = self.fh5[f"{sn}/primary/data/{list(self.det_name.values())[0]}"].shape[:-2]
-                if len(dshape)>1:
-                    if len(dshape)>2:
-                        raise Exception(f"Don't know how to handle data shape {dshape}")
-                    dshape = dshape[0]*dshape[1]
-                else:
-                    dshape = dshape[0]
-                if len(ts0) != dshape:
-                    raise Exception(f"mistached timestamp length: {len(ts0)} vs {dshape}")
-                if len(ts0)>1: # expect the monitor data to be 1D
-                    ts0 = ts0.flatten()
-            else:
-                raise Exception(f"timestamp data for {trigger} cannot be found.")
-
-            if strn=="primary":
                 trans_data0 = trans_data
                 incid_data0 = incid_data
             else:
+                # synch em1 time stamps to em2, LiX-specific 
+                synch_mon(ts2, trans_data, ts1, incid_data, dt=force_synch)
+                ts0 = self.get_ts(sn, exp, trigger)
+
                 try:
                     trans_data0 = integrate_mon(trans_data, ts2, ts0+force_synch_trig, exp, extend_mon_stream, **kwargs)
                     incid_data0 = integrate_mon(incid_data, ts1, ts0+force_synch_trig, exp, extend_mon_stream, **kwargs)   
@@ -978,11 +1098,16 @@ class h5xs():
                 self.d0s = {}
             if not sn in self.d0s.keys():
                 self.d0s[sn] = {}
-            self.d0s[sn]["transmitted"] = trans_data0
-            self.d0s[sn]["incident"] = incid_data0
+
+            # make sure that the monitor counts have the right number of data points
+            if check_size>0: 
+                check_size -= len(trans_data0)
+            self.d0s[sn]["transmitted"] = np.pad(trans_data0, (0,check_size), constant_values=np.nan)
+            self.d0s[sn]["incident"] = np.pad(incid_data0, (0,check_size), constant_values=np.nan)
             transmission = trans_data0/incid_data0
-            transmission /= np.nanmean(transmission)
-            self.d0s[sn]["transmission"] = transmission
+            #transmission /= np.nanmean(transmission)   # this can cause problems when the beam is off during part of the scan
+            self.d0s[sn]["transmission"] = np.pad(transmission, (0,check_size), constant_values=np.nan)
+            
             
     def set_trans(self, transMode, sn=None, trigger=None, gf_sigma=2, dt0=-133.8, exp=1, plot_trigger=False, **kwargs): 
         """ set the transmission values for the merged data
@@ -1046,6 +1171,7 @@ class h5xs():
             if 'averaged' in self.d1s[s].keys():
                 self.d1s[s]['averaged'].trans = np.average(t_values)
     
+    @h5_file_access  
     def load_d1s(self, sn=None):
         """ load the processed 1d data saved in the hdf5 file into memory 
             for each sample
@@ -1082,6 +1208,7 @@ class h5xs():
             self.d1s[sn][k] = unpack_d1(grp[k], self.qgrid, sn+k, tvs)   
             
 
+    @h5_file_access  
     def save_d1s(self, sn=None, debug=False):
         """
         save the 1d data in memory to the hdf5 file 
@@ -1116,13 +1243,13 @@ class h5xs():
             for k in list(self.attrs[sn].keys()):
                 grp.attrs[k] = self.attrs[sn][k]
                 if debug is True:
-                    print("writting attribute to %s: %s" % (sn, k))
+                    print(f"writing attribute to {sn}: {k}")
 
         ds_names = lsh5(grp, top_only=True, silent=True)
         for k in list(self.d1s[sn].keys()):
             data,tvs = pack_d1(self.d1s[sn][k])
             if debug is True:
-                print("writting attribute to %s: %s" % (sn, k))
+                print(f"writing attribute to {sn}: {k}")
             if k not in ds_names:
                 grp.create_dataset(k, data=data)
             else:
@@ -1277,6 +1404,7 @@ class h5xs():
                 self.d1s[sn]['averaged'].save("%s%s_%c.dat"%(path,sn,'a'), debug=debug, 
                                               footer=self.md_string(sn))
                 
+    @h5_file_access  
     def load_data(self, update_only=False, detectors=None,
            reft=-1, save_1d=False, save_merged=False, debug=False, N=8, max_c_size=0, dtype=None):
         """ assume multiple samples, parallel-process by sample
