@@ -24,6 +24,8 @@ det_model = {"_SAXS": "Pilatus3X 1M",
               "_WAXS1": "Pilatus3X 300K",
               "_WAXS2": "Pilatus3X 900K"}
 
+incident_monitor = "em1"
+transmitted_monitor = "em2"
 incident_monitor_field = "em1_sum_all_mean_value"
 transmitted_monitor_field = "em2_sum_all_mean_value"
 
